# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import os

import numpy as np
import paramiko
from plumbum import colors

import linien
from linien.client.exceptions import (
    InvalidServerVersionException,
    ServerNotInstalledException,
)
from linien.common import hash_username_and_password
from linien.config import REMOTE_BASE_PATH


def connect_ssh(host, user, password):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostname=host, username=user, password=password)
    return ssh


def run_server(host, user, password, port):
    ssh = connect_ssh(host, user, password)

    version = linien.__version__

    if version == "dev":
        # if we are in a development version, we upload the files from source
        # directory directly to the RP.
        upload_source_code(ssh)

        # start the development code that was just uploaded.
        # For that, go to the parent directory; an import of "linien" then
        # points to that directory instead of any globally installed release
        # version.
        stdin, stdout, stderr = ssh.exec_command(
            ("cd %s/../;" % REMOTE_BASE_PATH)
            + (
                "export LINIEN_AUTH_HASH=%s;"
                % hash_username_and_password(user, password)
            )
            + ("bash %s/server/linien_start_server.sh %d" % (REMOTE_BASE_PATH, port))
        )
        err = stderr.read()
        if err:
            print(colors.red | "Error starting the server")
            print(err)
    else:
        # it is a release and not a dev version.

        # read out version number of the server
        remote_version = read_remote_version(ssh)
        print("remote version", remote_version, "local version", version)

        if version != remote_version:
            raise InvalidServerVersionException(version, remote_version)

        # start the server process using the global command
        ssh.exec_command("linien_start_server.sh")
        ssh.close()


def read_remote_version(ssh):
    """Reads the remote version of linien using SSH."""
    stdin, stdout, stderr = ssh.exec_command(
        'python3 -c "import linien; print(linien.__version__);"'
    )
    lines = stdout.readlines()

    if not lines:
        raise ServerNotInstalledException()

    remote_version = lines[0].strip()
    return remote_version


def upload_source_code(ssh):
    """Uploads the application's source code to the remote server using SFTP."""
    print("uploading dev source code...")

    ftp = ssh.open_sftp()

    directory = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

    if not os.path.exists(os.path.join(directory, "server", "linien.bin")):
        print(
            colors.red | "Error: In order to run the development version, "
            "you need the FPGA bitstream in server/linien.bin! "
            "Consult README to see how you can get one."
        )
        raise Exception("FPGA bitstream missing")

    # upload the code required for running the server
    for dirpath, dirnames, filenames in os.walk(directory):
        # lstrip / so os.path.join does not think dir_path_rel is an absolute path.
        dirpath_rel = dirpath.replace(directory, "").lstrip("/")
        # Change direction of path slashes to work on the RedPitayas Linux system. This
        # is necessary when deploying the server from a Windows machine.
        dirpath_rel = dirpath_rel.lstrip("\\")
        remote_path = os.path.join(REMOTE_BASE_PATH, dirpath_rel)

        remote_path = remote_path.replace("\\", "/")

        if "." in dirpath_rel or "__" in dirpath_rel:
            continue

        try:
            ftp.lstat(remote_path)
        except IOError:
            ftp.mkdir(os.path.join(remote_path.rstrip("/")))

        for filename in filenames:
            local_path = os.path.join(dirpath, filename)
            remote_filepath = os.path.join(remote_path, filename)
            remote_filepath = remote_filepath.replace("\\", "/")
            # put file
            ftp.put(local_path, remote_filepath)

    ftp.close()


def peak_voltage_to_dBm(voltage):
    return 10 + 20 * np.log10(voltage)



import os

name = "linien"

from plumbum import colors

version_fn = os.path.join(
    *list(os.path.split(os.path.abspath(__file__))[:-1]) + ["VERSION"]
)

try:
    with open(version_fn, "r") as fh:
        __version__ = fh.read().strip()
        assert __version__ == "dev" or len(__version__.split(".")) == 3, (
            "invalid version number! Either has to be dev or something like " "1.0.1"
        )
except FileNotFoundError:
    print(
        colors.red | "Unable to read VERSION file. Create a file called "
        "VERSION in the checked_out_repo/linien folder. "
        'The content should be either "dev" for development or a '
        "version number if you want to release a new production version. "
        "Consult README for more details."
    )
    raise



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import random
import string
from socket import gaierror
from time import sleep
from traceback import print_exc
from typing import Callable

import rpyc
from plumbum import colors

import linien
from linien.client.exceptions import (
    GeneralConnectionErrorException,
    InvalidServerVersionException,
    RPYCAuthenticationException,
    ServerNotRunningException,
)
from linien.client.remote_parameters import RemoteParameters
from linien.client.utils import run_server

# IMPORTANT: keep this import, because it eases interfacing with the python client
from linien.common import ANALOG_OUT_V, MHz, Vpp, hash_username_and_password
from linien.config import DEFAULT_SERVER_PORT

assert MHz
assert Vpp
assert ANALOG_OUT_V


class RPYCClientWithAuthentication(rpyc.Service):
    """An rpyc client that authenticates using a hash.

    This class is run on the client side and exposes the client's unique id
    to the server."""

    def __init__(self, uuid, user, password):
        super().__init__()

        self.exposed_uuid = uuid
        self.auth_hash = hash_username_and_password(user, password).encode()

    def _connect(self, channel, config):
        # send auth hash before rpyc takes over
        channel.stream.sock.send(self.auth_hash)

        return super()._connect(channel, config)


class RawRPYCClient:
    """This class implements the basic functionality for connecting to a Linien
    server using rpyc. See `LinienClient` for higher-level functionality.

    Once connected, communication between server and client mainly takes place
    using the `parameters` attribute."""

    def __init__(
        self,
        server: str,
        port: int,
        user: str,
        password: str,
        use_parameter_cache: bool = False,
        call_on_error: Callable = None,
    ):
        self.use_parameter_cache = use_parameter_cache
        self.uuid = "".join(random.choice(string.ascii_lowercase) for i in range(10))

        # for exposing client's uuid to server
        self.client_service = RPYCClientWithAuthentication(self.uuid, user, password)

        self._connect(
            server,
            port,
            user,
            password,
            use_parameter_cache,
            call_on_error=call_on_error,
        )
        self.connected = True

    def _connect(
        self,
        server: str,
        port: int,
        user: str,
        password: str,
        use_parameter_cache: bool,
        call_on_error: Callable = None,
    ):
        """This method just redirects to `connect_rpyc` and is intended to be
        overridden in inheriting classes."""
        return self._connect_rpyc(
            server,
            port,
            user,
            password,
            use_parameter_cache,
            call_on_error=call_on_error,
        )

    def _connect_rpyc(
        self, server, port, user, password, use_parameter_cache, call_on_error=None
    ):
        """Connect to the server using rpyc and instanciate `RemoteParameters`."""
        self.connection = rpyc.connect(
            server, port, service=self.client_service, config={"allow_pickle": True}
        )

        cls = RemoteParameters
        if call_on_error:
            cls = self._catch_network_errors(cls, call_on_error)

        self.parameters = cls(self.connection.root, self.uuid, use_parameter_cache)

    def _catch_network_errors(self, cls, call_on_error):
        """This method can be used for patching RemoteParameters such
        that network errors are redirected to `call_on_error`"""
        function_type = type(lambda x: x)

        for attr_name in dir(cls):
            # patch all methods that don't start with __
            if not attr_name.startswith("__"):
                attr = getattr(cls, attr_name)

                if isinstance(attr, function_type):
                    method = attr

                    def wrapped(*args, method=method, **kwargs):
                        try:
                            return method(*args, **kwargs)
                        except (EOFError,):
                            print(colors.red | "Connection lost")
                            self.connected = False
                            call_on_error()
                            raise

                    setattr(cls, attr_name, wrapped)

        return cls


class LinienClient(RawRPYCClient):
    def __init__(
        self,
        device,
        autostart_server=True,
        use_parameter_cache=False,
        on_connection_lost: Callable = None,
    ):
        """Connect to a RedPitaya that runs linien server.

        Takes the following arguments:
            * `device` should be a dictionary:
                {
                    "host": "rp-XXXXXX.local",
                    "username": "root",
                    "password": "your-username"
                }
            * `autostart_server`: A bool indicating whether this call
              should automatically start a linien server on redpitaya if it
              doesn't run already
        """
        self.device = device
        self.host = device["host"]
        user = device.get("username")
        password = device.get("password")

        if self.host in ("localhost", "127.0.0.1"):
            # RP is configured such that "localhost" doesn't point to
            # 127.0.0.1 in all cases
            self.host = "127.0.0.1"
        else:
            assert user and password, "username and passwort are required"

        self.autostart_server = autostart_server

        super().__init__(
            self.host,
            device.get("port", DEFAULT_SERVER_PORT),
            user,
            password,
            use_parameter_cache=use_parameter_cache,
            call_on_error=on_connection_lost,
        )

    def _connect(
        self, host, port, user, password, use_parameter_cache, call_on_error=None
    ):
        self.connection = None

        i = -1
        while True:
            i += 1

            try:
                print("try to connect to %s:%s" % (host, port))
                self._connect_rpyc(
                    host,
                    port,
                    user,
                    password,
                    use_parameter_cache,
                    call_on_error=call_on_error,
                )
                self.control = self.connection.root
                break
            except gaierror:
                # host not found
                print(colors.red | "Error: host %s not found" % host)
                break
            except EOFError:
                print("EOFError! Probably authentication failed")
                raise RPYCAuthenticationException()
            except Exception:
                if not self.autostart_server:
                    raise ServerNotRunningException()

                if i == 0:
                    print("server is not running. Launching it!")
                    run_server(host, user, password, port)
                    sleep(3)
                else:
                    if i < 20:
                        print(
                            """
                            Server still not running, waiting (this may take some time).
                            """
                        )
                        sleep(1)
                    else:
                        print_exc()
                        print(
                            colors.red
                            | "Error: connection to the server could not be established"
                        )
                        break

        if self.connection is None:
            raise GeneralConnectionErrorException()

        # now check that the remote version is the same as ours
        remote_version = self.connection.root.exposed_get_server_version()
        client_version = linien.__version__

        if remote_version != client_version:
            raise InvalidServerVersionException(client_version, remote_version)

        print(colors.green | "connected established!")

    def disconnect(self):
        self.connection.close()
        self.connected = False



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable

import rpyc
from rpyc import async_

from linien.common import pack, unpack


class RemoteParameters:
    """A class that provides access to a remote `parameters.Parameters` instance.

    It clones the functionality of the remote `Parameters` instance. E.g.:

        # on the remote side
        p = Parameters(...)
        p.my_param.value = 123

        # on the client side
        r = RemoteParameters(...)

        # RemoteParameters allows for accessing the remote value:
        print(r.my_param.value) # outputs 123

        # it's also possible to set the value (this change is automatically
        # propagated to the server)
        r.my_param.value = 123

        # and we can set up a callback function that is called whenever a
        # parameter changes
        def on_change(value):
            # this function is called whenever `my_param` changes on the server.
            # note that this only works if `call_listeners` is called from
            # time to time as this function is responsible for checking for
            # changed parameters.
            print('parameter arrived!', value)
        r.my_param.on_change(on_change)
        while True:
            r.call_listeners()
            sleep(.1)

    The arguments for __init__ are:

        `remote`:    The root of the rpyc connection ot the server
        `uuid`:      A random unique identifier for this client
        `use_cache`: A boolean indicating whether (most) parameters should be
                     cached locally. If this is not enabled, every access of
                     `r.my_param.value` results in a request to the server.
                     If `use_cache` is enabled, a local cache is used instead.
                     For that purpose, a listener is installed such that the
                     server automatically pushes changes to the client and
                     thus updates the cache. No matter how often you access
                     `r.my_param.value`, each parameter value is only transmitted
                     once (after it was changed).
                     Note that calling `call_listeners` is required for this.
    """

    def __init__(self, remote, uuid: str, use_cache: bool):
        self.remote = remote
        self.uuid = uuid
        self._async_listener_queue = None
        self._async_listener_registering = None

        self._listeners_pending_remote_registration = []
        self._listeners = {}

        self._mimic_remote_parameters(use_cache)
        self._attributes_locked = True

        self.call_listeners()

    def __setattr__(self, name, value):
        """In order to set the value of a parameter,

            parameters.my_param.value = 123

        is used. In order to prevent accidentally forgetting the .value part, i.e.

            parameters.my_param = 123

        we raise an error in this case."""
        if (
            hasattr(self, "_attributes_locked")
            and self._attributes_locked
            and not name.startswith("_")
        ):
            raise Exception(
                "Parameters are locked! Did you mean to set the value of this "
                "parameter instead, i.e. parameters.%s.value = %s" % (name, value)
            )
        super().__setattr__(name, value)

    def _mimic_remote_parameters(self, use_cache: bool):
        """For every remote parameter, instanciate a `RemoteParameter` object
        that allows to mimics the functionality of the remote parameter."""
        # when directly iterating over `exposed_init_parameter_sync`, each iteration
        # triggers a request as it is a netref over an iterator
        # --> the `list` call prevents this and improves startup performance
        all_parameters = unpack(self.remote.exposed_init_parameter_sync(self.uuid))

        for name, param, value, can_be_cached in all_parameters:
            param = RemoteParameter(self, param, name, use_cache and can_be_cached)
            setattr(self, name, param)
            if use_cache and can_be_cached:
                # obtain takes care that we really don't deal with netrefs
                # (np.float64 is not automatically serialized)
                param._update_cache(rpyc.classic.obtain(value))

        self._attributes_locked = True

    def _register_listener(self, param, callback: Callable):
        """Tells the server to notify our client (identified by `self.uuid`)
        when `param` changes. Registers a function `callback` that will be
        called in this case."""
        if param.name not in self._listeners:
            # parameters that use the cache don't have to be registered on remote
            # side because the client automatically listens for changes.
            # This happens using `init_parameter_sync`
            if not param.use_cache:
                self._listeners_pending_remote_registration.append(param.name)

        self._listeners.setdefault(param.name, [])
        self._listeners[param.name].append(callback)

    def call_listeners(self):
        """Ask the server for changed parameters and call the respective
        callback functions. This call takes place asynchronously, i.e. the first
        run of `call_listeners` just issues the call but does not wait for it
        in order not to block the GUI. The following calls check whether a
        result has arrived (also not blocking GUI).

        In Linien GUI client, this function is called periodically.
        If you use the python client and want to use callbacks for changed
        parameters you have to call this method manually from time to time."""

        def _get_listener_queue_async():
            """Issues an asynchronous call (that does not block the GUI) to the
            server in order to retrieve a batch of changed parameters."""
            self._async_listener_queue = async_(self.remote.get_listener_queue)(
                self.uuid
            )

        def _register_listeners_async():
            """Issues an asynchronous call to the server containing all the
            parameters that this client wants to be notified of in case of a
            changed value."""
            pending = self._listeners_pending_remote_registration
            if pending:
                # this copies the list before clearing it below. Otherwise
                # we just transmit an epmty list in the async call
                pending = pending[:]
                self._async_listener_registering = async_(
                    self.remote.exposed_register_remote_listeners
                )(self.uuid, pending)
                self._listeners_pending_remote_registration.clear()

        if self._async_listener_queue is None:
            # this means that the async call was not started yet --> start it
            # the next call to `call_listeners` will then check whether the
            # the result is ready.
            _get_listener_queue_async()

        if self._async_listener_registering is None:
            _register_listeners_async()

        if self._async_listener_queue is not None and self._async_listener_queue.ready:
            # we have a result
            queue = unpack(self._async_listener_queue.value)

            # now that we have our result, we can start the next call
            _get_listener_queue_async()

            # before calling listeners, we update cache for all received parameters
            # at once
            for param_name, value in queue:
                param = getattr(self, param_name)
                if param.use_cache:
                    param._update_cache(value)

            # iterate over all canged parameters and call the respective
            # callback functions
            for param_name, value in queue:
                if param_name in self._listeners:
                    for listener in self._listeners[param_name]:
                        listener(value)

        if (
            self._async_listener_registering is not None
            and self._async_listener_registering.ready
        ):
            # registration of listeners was successful on the remote side
            # now we can clear the async call object such that a new one may
            # be issued (if required)
            self._async_listener_registering = None

    def _get_param(self, param_name):
        return unpack(self.remote.exposed_get_param(param_name))

    def _set_param(self, param_name, value):
        return self.remote.exposed_set_param(param_name, pack(value))


class RemoteParameter:
    """A helper class for `RemoteParameters`, representing a single remote
    parameter."""

    def __init__(
        self, parent: RemoteParameters, remote_param, name: str, use_cache: bool
    ):
        self._remote_param = remote_param
        self.name = name
        self.parent = parent
        self.use_cache = use_cache

    @property
    def value(self):
        """Return the locally cached value (if it exists). Otherwise ask the
        server."""
        if hasattr(self, "_cached_value"):
            return self._cached_value
        return self.parent._get_param(self.name)

    @value.setter
    def value(self, value):
        """Notify the server of the new value"""
        return self.parent._set_param(self.name, value)

    def on_change(self, callback_on_change, call_listener_with_first_value=True):
        """Tells the server that `callback_on_change` should be called whenever
        the parameter changes."""
        self.parent._register_listener(self, callback_on_change)

        if call_listener_with_first_value:
            # call the callback with the initial value
            callback_on_change(self.value)

    def reset(self):
        """Reset the value to its initial value"""
        self._remote_param.reset()

    def _update_cache(self, value):
        self._cached_value = value



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import setuptools

import linien

assert linien.__version__ != "dev"

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="linien-python-client",
    version=linien.__version__,
    author="Benjamin Wiegand",
    author_email="highwaychile@posteo.de",
    maintainer="Bastian Leykauf",
    maintainer_email="leykauf@physik.hu-berlin.de",
    description="Python client for linien spectroscopy lock",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/linien-org/linien/",
    # IMPORTANT: any changes have to be made in setup_client_and_gui.py
    # of flathub repo as well
    packages=["linien", "linien.client"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy>=1.19.1",
        "paramiko>=2.10.1",
        "plumbum>=1.6.9",
        "rpyc>=4.0,<5.0",
        "scipy>=1.4.1",
        "uuid>=1.30",
    ],
    package_data={
        # IMPORTANT: any changes have to be made in pyinstaller.spec, too
        # (for the standalone installer)
        # IMPORTANT: any changes have to be made in setup_client_and_gui.py
        # of flathub repo as well
        "": ["VERSION"]
    },
)
